@@ -187,6 +187,8 @@ static void on_action_directory_previous       (GtkAction * action,
 						GeditFileBrowserWidget * obj);
 static void on_action_directory_next           (GtkAction * action,
 						GeditFileBrowserWidget * obj);
+static void on_action_directory_up             (GtkAction * action,
+						GeditFileBrowserWidget * obj);
 static void on_action_directory_new            (GtkAction * action,
 						GeditFileBrowserWidget * obj);
 static void on_action_file_new                 (GtkAction * action,
@@ -518,13 +520,17 @@ get_from_bookmark (GeditFileBrowserWidget * obj, gchar const *uri,
 		   gchar ** name, GdkPixbuf ** icon)
 {
 	GnomeVFSURI *guri;
+	gboolean result;
 
 	guri = gnome_vfs_uri_new (uri);
 
 	if (guri == NULL)
 		return FALSE;
 	
-	return get_from_bookmark_uri (obj, guri, name, icon);
+	result = get_from_bookmark_uri (obj, guri, name, icon);
+	gnome_vfs_uri_unref (guri);
+	
+	return result;
 }
 
 static void
@@ -600,16 +606,24 @@ static guint
 uri_num_parents (GnomeVFSURI * from, GnomeVFSURI * to)
 {
 	guint parents = 0;
-	
+	GnomeVFSURI * tmp;
+
 	if (from == NULL)
 		return 0;
 	
+	from = gnome_vfs_uri_dup (from);
+	
 	while (gnome_vfs_uri_has_parent (from) && 
 	       !(to && gnome_vfs_uri_equal (from, to))) {
-		from = gnome_vfs_uri_get_parent (from);
+		tmp = gnome_vfs_uri_get_parent (from);
+		gnome_vfs_uri_unref (from);
+		from = tmp;
+		
 		++parents;
 	}
 	
+	gnome_vfs_uri_unref (from);
+	
 	return parents;
 }
 
@@ -620,6 +634,7 @@ insert_location_path (GeditFileBrowserWidget * obj)
 	GnomeVFSURI *virtual;
 	GnomeVFSURI *root;
 	GnomeVFSURI *current = NULL;
+	GnomeVFSURI * tmp;
 	GtkTreeIter separator;
 	GtkTreeIter iter;
 	guint indent;
@@ -655,13 +670,18 @@ insert_location_path (GeditFileBrowserWidget * obj)
 							   obj);
 		}
 
-		if (gnome_vfs_uri_equal (current, root))
+		if (gnome_vfs_uri_equal (current, root) || !gnome_vfs_uri_has_parent (current)) {
+			if (current != virtual)
+				gnome_vfs_uri_unref (current);
 			break;
+		}
 
-		if (!gnome_vfs_uri_has_parent (current))
-			break;
+		tmp = gnome_vfs_uri_get_parent (current);
+		
+		if (current != virtual)
+			gnome_vfs_uri_unref (current);
 
-		current = gnome_vfs_uri_get_parent (current);
+		current = tmp;
 	}
 
 	gnome_vfs_uri_unref (virtual);
@@ -711,20 +731,6 @@ fill_combo_model (GeditFileBrowserWidget * obj)
 	gtk_combo_box_set_active (GTK_COMBO_BOX (obj->priv->combo), 0);
 }
 
-static void
-cell_data_func (GtkCellLayout * cell_layout, GtkCellRenderer * cell,
-		GtkTreeModel * model, GtkTreeIter * iter, gpointer data)
-{
-	/*guint id;
-
-	   gtk_tree_model_get(model, iter, COLUMN_ID, &id, -1);
-
-	   if (id == PATH_ID)
-	   g_object_set(cell, "weight", PANGO_WEIGHT_BOLD, "mode", GTK_CELL_RENDERER_MODE_INERT, NULL);
-	   else
-	   g_object_set(cell, "weight", PANGO_WEIGHT_NORMAL, "mode", GTK_CELL_RENDERER_MODE_ACTIVATABLE, NULL); */
-}
-
 static void
 indent_cell_data_func (GtkCellLayout * cell_layout, 
                        GtkCellRenderer * cell,
@@ -782,10 +788,6 @@ create_combo (GeditFileBrowserWidget * obj)
 	gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (obj->priv->combo),
 				       renderer, "text", COLUMN_NAME);
 
-	gtk_cell_layout_set_cell_data_func (GTK_CELL_LAYOUT
-					    (obj->priv->combo), renderer,
-					    cell_data_func, obj, NULL);
-
 	gtk_box_pack_start (GTK_BOX (obj), GTK_WIDGET (obj->priv->combo),
 			    FALSE, FALSE, 0);
 
@@ -807,7 +809,9 @@ static GtkActionEntry tree_actions[] =
 	 N_("Add new empty directory"),
 	 G_CALLBACK (on_action_directory_new)},
 	{"FileNew", GTK_STOCK_NEW, N_("New F_ile"), NULL,
-	 N_("Add new empty file"), G_CALLBACK (on_action_file_new)}
+	 N_("Add new empty file"), G_CALLBACK (on_action_file_new)},
+	{"DirectoryUp", GTK_STOCK_GO_UP, N_("Up"), NULL,
+	 N_("Open the parent folder"), G_CALLBACK (on_action_directory_up)}
 };
 
 static GtkActionEntry tree_actions_selection[] = 
@@ -827,10 +831,10 @@ static GtkActionEntry tree_actions_sensitive[] =
 {
 	{"DirectoryPrevious", GTK_STOCK_GO_BACK, N_("_Previous Location"),
 	 NULL,
-	 N_("Go to previous location"),
+	 N_("Go to the previous visited location"),
 	 G_CALLBACK (on_action_directory_previous)},
 	{"DirectoryNext", GTK_STOCK_GO_FORWARD, N_("_Next Location"), NULL,
-	 N_("Go to next location"), G_CALLBACK (on_action_directory_next)},
+	 N_("Go to the next visited location"), G_CALLBACK (on_action_directory_next)},
 	{"DirectoryRefresh", GTK_STOCK_REFRESH, N_("Re_fresh View"), NULL,
 	 N_("Refresh the view"), G_CALLBACK (on_action_directory_refresh)},
 	{"DirectoryOpen", GTK_STOCK_OPEN, N_("_View Directory"), NULL,
@@ -979,13 +983,6 @@ create_toolbar (GeditFileBrowserWidget * obj)
 	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), GTK_TOOL_ITEM (widget),
 			    1);
 
-	/* Separator tool item */
-	widget = GTK_WIDGET (gtk_separator_tool_item_new ());
-	gtk_widget_show (widget);
-
-	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), GTK_TOOL_ITEM (widget),
-			    2);
-
 	gtk_box_pack_start (GTK_BOX (obj), toolbar, FALSE, FALSE, 0);
 	gtk_widget_show (toolbar);
 }
@@ -1364,13 +1361,21 @@ delete_selected_file (GeditFileBrowserWidget * obj, gboolean trash)
 	return result == GEDIT_FILE_BROWSER_STORE_RESULT_OK;
 }
 
-static GnomeVFSURI const *
+static GnomeVFSURI *
 get_topmost_uri (GnomeVFSURI const *uri)
 {
-	while (gnome_vfs_uri_has_parent (uri))
-		uri = gnome_vfs_uri_get_parent (uri);
-
-	return uri;
+	GnomeVFSURI * tmp;
+	GnomeVFSURI * current;
+	
+	current = gnome_vfs_uri_dup (uri);
+	
+	while (gnome_vfs_uri_has_parent (current)) {
+		tmp = gnome_vfs_uri_get_parent (current);
+		gnome_vfs_uri_unref (current);
+		current = tmp;
+	}
+	
+	return current;
 }
 
 static GtkWidget *
@@ -1387,7 +1392,7 @@ create_goto_menu_item (GeditFileBrowserWidget * obj, GList * item,
 
 	if (!get_from_bookmark
 	    (obj, loc->virtual_root, &unescape, &pixbuf)) {
-		if (strncmp (loc->virtual_root, "file://", 7) == 0) {
+		if (gedit_utils_uri_has_file_scheme (loc->virtual_root)) {
 			unescape =
 			    gnome_vfs_get_local_path_from_uri (loc->
 							       virtual_root);
@@ -1702,7 +1707,7 @@ gedit_file_browser_widget_set_root (GeditFileBrowserWidget * obj,
 				    gboolean virtual)
 {
 	GnomeVFSURI *uri;
-	GnomeVFSURI const *parent;
+	GnomeVFSURI *parent;
 	gchar *str;
 
 	if (!virtual) {
@@ -1723,6 +1728,7 @@ gedit_file_browser_widget_set_root (GeditFileBrowserWidget * obj,
 
 			g_free (str);
 			gnome_vfs_uri_unref (uri);
+			gnome_vfs_uri_unref (parent);
 		} else {
 			str =
 			    g_strconcat (_("Invalid uri"), ": ", root,
@@ -1889,6 +1895,22 @@ on_end_loading (GeditFileBrowserStore * model, GtkTreeIter * iter,
 				       NULL);
 }
 
+static gboolean
+virtual_root_is_root (GeditFileBrowserWidget * obj, 
+                      GeditFileBrowserStore  * model)
+{
+	GtkTreeIter root;
+	GtkTreeIter virtual;
+	
+	if (!gedit_file_browser_store_get_iter_root (model, &root))
+		return TRUE;
+	
+	if (!gedit_file_browser_store_get_iter_virtual_root (model, &virtual))
+		return TRUE;
+	
+	return gedit_file_browser_store_iter_equal (model, &root, &virtual);
+}
+
 static void
 on_virtual_root_changed (GeditFileBrowserStore * model,
 			 GParamSpec * param,
@@ -1956,6 +1978,13 @@ on_virtual_root_changed (GeditFileBrowserStore * model,
 					g_object_unref (pixbuf);
 
 			}
+			
+			action =
+			    gtk_action_group_get_action (obj->priv->
+			                                 action_group,
+			                                 "DirectoryUp");
+			gtk_action_set_sensitive (action, 
+			                          !virtual_root_is_root (obj, model));
 
 			action =
 			    gtk_action_group_get_action (obj->priv->
@@ -1996,14 +2025,14 @@ on_model_set (GObject * gobject, GParamSpec * arg1,
 
 	clear_signals (obj);
 
-	if (GEDIT_IS_FILE_BOOKMARKS_STORE (model)) {
-		add_signal (obj, gobject,
-			    g_signal_connect_after (gobject,
-						    "row-activated",
-						    G_CALLBACK
-						    (on_row_activated),
-						    obj));
+	add_signal (obj, gobject,
+		    g_signal_connect (gobject,
+				      "row-activated",
+				      G_CALLBACK
+				      (on_row_activated),
+				      obj));
 
+	if (GEDIT_IS_FILE_BOOKMARKS_STORE (model)) {
 		clear_next_locations (obj);
 
 		/* Add the current location to the back menu */
@@ -2027,12 +2056,6 @@ on_model_set (GObject * gobject, GParamSpec * arg1,
 
 		gtk_widget_set_sensitive (obj->priv->filter_vbox, FALSE);
 	} else if (GEDIT_IS_FILE_BROWSER_STORE (model)) {
-		add_signal (obj, gobject,
-			    g_signal_connect_after (gobject,
-						    "row-activated",
-						    G_CALLBACK
-						    (on_row_activated),
-						    obj));
 		add_signal (obj, model,
 			    g_signal_connect (model, "begin-loading",
 					      G_CALLBACK
@@ -2126,11 +2149,59 @@ on_treeview_button_press_event (GeditFileBrowserView * treeview,
 	return FALSE;
 }
 
+static gboolean
+do_change_directory (GeditFileBrowserWidget * obj,
+                     GdkEventKey            * event)
+{
+	GtkAction * action = NULL;
+
+	if ((event->state & 
+	    (~GDK_CONTROL_MASK & ~GDK_SHIFT_MASK & ~GDK_MOD1_MASK)) == 
+	     event->state && event->keyval == GDK_BackSpace)
+		action = gtk_action_group_get_action (obj->priv->
+		                                      action_group_sensitive,
+		                                      "DirectoryPrevious");
+	else if (!((event->state & GDK_MOD1_MASK) && 
+	    (event->state & (~GDK_CONTROL_MASK & ~GDK_SHIFT_MASK)) == event->state))
+		return FALSE;
+
+	switch (event->keyval) {
+		case GDK_Left:
+			action = gtk_action_group_get_action (obj->priv->
+			                                      action_group_sensitive,
+			                                      "DirectoryPrevious");
+		break;
+		case GDK_Right:
+			action = gtk_action_group_get_action (obj->priv->
+			                                      action_group_sensitive,
+			                                      "DirectoryNext");
+		break;
+		case GDK_Up:
+			action = gtk_action_group_get_action (obj->priv->
+			                                      action_group,
+			                                      "DirectoryUp");
+		break;
+		default:
+		break;
+	}
+	
+	if (action != NULL) {
+		gtk_action_activate (action);
+		return TRUE;
+	}
+	
+	return FALSE;
+}
+
+	
 static gboolean
 on_treeview_key_press_event (GeditFileBrowserView * treeview,
 			     GdkEventKey * event,
 			     GeditFileBrowserWidget * obj)
 {
+	if (do_change_directory (obj, event))
+		return TRUE;
+
 	if (!GEDIT_IS_FILE_BROWSER_STORE
 	    (gtk_tree_view_get_model (GTK_TREE_VIEW (treeview))))
 		return FALSE;
@@ -2268,6 +2339,20 @@ on_action_directory_previous (GtkAction * action,
 	}
 }
 
+static void 
+on_action_directory_up (GtkAction              * action,
+			GeditFileBrowserWidget * obj)
+{
+	GtkTreeModel *model =
+	    gtk_tree_view_get_model (GTK_TREE_VIEW (obj->priv->treeview));
+	   
+	if (!GEDIT_IS_FILE_BROWSER_STORE (model))
+		return;
+	
+	gedit_file_browser_store_set_virtual_root_up (GEDIT_FILE_BROWSER_STORE (model));	
+}
+
+
 static void
 on_action_directory_new (GtkAction * action, GeditFileBrowserWidget * obj)
 {
